module vars

! allocatable arrays and namelist variables
  
  implicit none

! --------------------------- coordinate variables ---------------------------

  real, allocatable, dimension(:) ::                                          &
       sigma,              &   ! vertical coordinate axis (full levels)
       deg_lat,            &   ! latitude in degrees [local_num_lat]
       sin_lat,            &   ! sin(latitude)
       theta_coord,        &   ! theta coodinate axis (degrees)
       dtheta_coord,       &   ! spacing between theta levels (degrees)
       pk,                 &   ! pressure hybrid coefficients
       rhum_bin,           &   ! relative humidity pdf bin axis
       bk                      ! sigma hybrid coefficients

  real, allocatable, dimension(:,:) ::                                        &
       cos_lat_xy,         &   ! xy cosine(latitude) array
       cos_lat_yz              ! yz cosine(latitude) array

!------------------ gridpoint variables on pressure surfaces -------------------

  real, allocatable, dimension(:,:,:) ::                                      &
       u_grid,             &   ! zonal wind
       shum_grid,          &   ! specific humidity
       dt_qg_cond_grid,    &   ! large scale cond specific humidity tendency
       dt_qg_conv_grid,    &   ! convection specific humidity tendency
       dt_qg_cond_counter, &   ! counter for large scale cond specific humidity tendency
       dt_qg_conv_counter, &   ! counter for convection specific humidity tendency
       ozone_grid,         &   ! ozone
       v_grid,             &   ! meridional wind
       vor_grid,           &   ! vorticity
       div_grid,           &   ! divergence
       temp_grid,          &   ! temperature
       virtual_temp_grid,  &   ! virtual temperature
       vcos_grid,          &   ! cos(lat) * v
       w_grid,             &   ! vertical velocity (full levels)
       w_up_grid,          &   ! upward vertical velocity d(sigma)/dt
       w_up_freq_grid,     &   ! frequency of upward vertical velocity
       p_half,             &   ! pressure on half levels
       p_full,             &   ! pressure on full levels
       sfctn_grid,         &   ! streamfunction
       hor_sfctn_grid,     &   ! horizontal streamfunction
       pot_temp,           &   ! potential temperature
       theta_e,            &   ! equivalent potential temperature
       theta_e_sat,        &   ! saturated equivalent potential temperature
       geopot_grid,        &   ! geopotential
       mont_grid,          &   ! Montgomery streamfunction
       mont_x_grid,        &   ! x derivative of Montgomery streamfunction
       sat_mr_grid,        &   ! saturation mixing ratio
       sat_shum_grid,      &   ! saturation specific humidity
       rhum_grid               ! relative humidity




  real, allocatable, dimension(:,:)  ::                                       &
       ps_grid,            &   ! surface pressure
       surf_montgomery,    &   ! surface Montgomery streamfunction
       dx_surf_montgomery, &   ! derivative of surf. Mont. sf. WRT longitude
       dy_surf_montgomery, &   ! derivative of surf. Mont. sf. WRT longitude
       surf_bernoulli,     &   ! surface Bernoulli function
       dx_surf_pottemp,    &   ! derivative of surf. pot. temp. WRT longitude
       dy_surf_pottemp,    &   ! derivative of surf. pot. temp. WRT latitude
       surf_geopotential,  &   ! surface geopotential
       ucos_zon_spec_avg,  &   ! u*cos(lat) as func. of wavenumber (vert. avg.)
       vcos_zon_spec_avg,  &   ! v*cos(lat) as func. of wavenumber (vert. avg.)
       shum_zon_spec_avg,  &   ! shum as func. of wavenumber (vert. avg.)
       ozone_zon_spec_avg, &   ! ozone as func. of wavenumber (vert. avg.)
       spv_flux_y,         &   ! y component of surface PV flux
       bt_sfctn_grid,      &   ! barotropic streamfunction
       precip_cond_grid,   &   ! large scale cond precipitation
       precip_conv_grid,   &   ! convection precipitation
       precip_daily,       &   ! daily precipitation 
       precip_cond_daily,  &   ! daily large scale cond precipitation 
       precip_conv_daily,  &   ! daily convection precipitation 
       one_array               ! array of ones

  logical, allocatable, dimension(:,:)  ::                                       &
       precip_mask             ! mask for determining wet days

  real, allocatable, dimension(:, :, :) ::                                    &
       pot_temp_spec,      &   ! potential temperature spectrum (all levels)
       theta_e_spec,       &   ! equivalent potential temperature spectrum (all levels)
       shum_spec,          &   ! specific humidity spectrum (all levels)
       sat_shum_spec,      &   ! sat specific humidity spectrum (all levels)
       rhum_spec,          &   ! relative humidity spectrum (all levels)
       ozone_spec,         &   ! ozone spectrum (all levels)
       conv_spec               ! conv pe->ke spectrum (all levels)

  real, allocatable, dimension(:, :, :) ::                                    &
       rhum_pdf                ! pdf of relative humidity (lat, level, bin)
  
  real, allocatable, dimension(:, :) ::                                       &
       bc_sfctn_spec,      &   ! baroclinic energy spectrum
       bt_sfctn_spec           ! barotropic energy spectrum

       
!------------------- gridpoint variables on theta surfaces ---------------------

  real, allocatable, dimension(:, :, :) ::                                    &
       p_isentr,           &   ! pressure
       dens_isentr,        &   ! density
       ipv,                &   ! isentropic potential vorticity (PV)
       v_isentr,           &   ! meridional wind
       u_isentr,           &   ! zonal wind
       shum_isentr,        &   ! specific humidity
       ozone_isentr,       &   ! ozone
       rhum_isentr,        &   ! relative humidity
       sat_shum_isentr,    &   ! saturation specific humidity
       vshum_isentr,       &   ! specific humidity * meridional velocity
       vozone_isentr,      &   ! ozone * meridional velocity
       vsat_shum_isentr,   &   ! saturation specific humidity * meridional velocity
       abs_vor_isentr,     &   ! absolute vorticity
       sfctn_isentr,       &   ! streamfunction
       mont_isentr,        &   ! montgomery streamfunction
       mont_x_isentr,      &   ! x derivative of montgomery streamfunction
       rho_mont_x_isentr       ! isentropic density * mont_x_isentr


  integer, allocatable, dimension(:, :) ::                                    &
       k_sfc,              &   ! index of surface potential temperature
       k_asfc                  ! index of lowest isentropic layer above surface

! --------- zonally averaged gridpoint variables on pressure surfaces ---------

  real, allocatable, dimension(:, :) ::                                       &
       p_half_zon,               &   ! pressure on half levels
       p_full_avg,               &   ! pressure on full levels
       z_avg,                    &   ! geopotential height
       temp_avg,                 &   ! temperature
       virtual_temp_avg,         &   ! virtual temperature
       temp_var_avg,             &   ! temperature variance
       pot_temp_avg,             &   ! potential temperature
       pot_temp_var_avg,         &   ! potential temperature variance
       theta_e_avg,              &   ! equivalent potential temperature
       theta_e_sat_avg,          &   ! saturated equivalent potential temperature
       theta_e_var_avg,          &   ! equivalent potential temperature variance
       u_avg,                    &   ! zonal wind
       shum_avg,                 &   ! specific humidity
       dt_qg_cond_avg,           &   ! large scale cond specific humidity tendency
       dt_qg_conv_avg,           &   ! convection specific humidity tendency
       dt_qg_cond_freq,          &   ! large scale cond specific humidity tendency frequency
       dt_qg_conv_freq,          &   ! convection specific humidity tendency frequency
       ozone_avg,                &   ! ozone
       shum_var_avg,             &   ! specific humidity variance
       sat_shum_var_avg,         &   ! saturation specific humidity variance
       sat_shum_avg,             &   ! saturation specific humidity
       rhum_avg,                 &   ! relative humidity
       vshum_avg,                &   ! specific humidity * meridional velocity
       vozone_avg,               &   ! ozone * meridional velocity
       wshum_avg,                &   ! specific humidity * vertical velocity (\dot\sigma)
       temp_shum_avg,            &   ! eddy correlation of specific humidity and temperature
       u_var_avg,                &   ! zonal wind variance
       v_avg,                    &   ! meridional wind
       vort_avg,                 &   ! meridional wind
       vcos_avg,                 &   ! cosine(latitude)*meridional wind
       w_avg,                    &   ! vertical wind (w =  \dot\sigma)
       w_up_avg,                 &   ! upward component of vertical wind
       w_up_freq,                &   ! avg frequency of upward vert. wind
       v_var_avg,                &   ! meridional wind variance
       w_var_avg,                &   ! vertical wind variance
       w_cube_avg,               &   ! vertical wind eddy cube
       uvcos_avg,                &   ! u*cos(latitude)*v
       wu_avg,                   &   ! w*u ; w =  \dot\sigma
       uvcos_eddy,               &   ! u*cos(latitude)*v
       wu_eddy,                  &   ! w*u ; w =  \dot\sigma
       svol_avg,                 &   ! specific volume
       d_dy_pot_temp_avg,        &   ! derivative of pot. temp. WRT latitude
       d_dy_theta_e_avg,         &   ! derivative of equiv. pot. temp. WRT latitude
       mrdnl_z_flux_avg,         &   ! zonal mean [ psfc * v * cos(lat) * z ]
       vrtcl_z_flux_avg,         &   ! vertical z flux 
       mrdnl_heatflux_avg,       &   ! zonal mean [ psfc * v * cos(lat) * pot_temp ]
       mrdnl_theta_e_flux_avg,   &   ! zonal mean [ psfc * v * cos(lat) * theta_e ]
       vrtcl_heatflux_avg,       &   ! vertical heat flux
       vrtcl_theta_e_flux_avg,   &   ! vertical equiv. pot. temperature flux
       mrdnl_tempflux_avg,       &   ! zonal mean [ psfc * v * cos(lat) * temp ]
       vrtcl_tempflux_avg,       &   ! vertical temperature flux
       mrdnl_eddy_z_flux,        &   ! meridional eddy z flux
       vrtcl_eddy_z_flux,        &   ! vertical eddy z flux
       mrdnl_eddy_heatflux,      &   ! meridional eddy heat flux
       mrdnl_eddy_theta_e_flux,  &   ! meridional eddy equiv. pot. temp flux
       vrtcl_eddy_heatflux,      &   ! vertical eddy heat flux
       vrtcl_eddy_theta_e_flux,  &   ! vertical equiv. pot. temp flux
       mrdnl_eddy_tempflux,      &   ! meridional eddy temperature flux
       vrtcl_eddy_tempflux,      &   ! vertical eddy temperature flux
       eddy_temp_div,            &   ! divergence of eddy temperature flux
       sfctn_avg,                &   ! streamfunction
       hor_sfctn_avg,            &   ! horizontal streamfunction
       hor_sfctn_var_avg,        &   ! horizontal streamfunction variance
       d_dsig_pot_temp_avg,      &   ! derivative of pot. temp. WRT sigma
       d_dsig_theta_e_avg,       &   ! derivative of equiv. pot. temp. WRT sigma
       d_dp_pot_temp_avg,        &   ! derivative of pot. temp. WRT pressure
       d_dp_theta_e_avg,         &   ! derivative of equiv. pot. temp. WRT pressure
       freq_stat_unstable,       &   ! frequency of statically unstable conditions
       TEM_res_circ,             &   ! TEM residual circulation
       MTEM_res_circ,            &   ! Modified TEM residual circulation
       buoyancy_avg                  ! Brunt-Vaeisaelae frequency

  real, allocatable, dimension(:) ::                                          &
       ps_grid_avg,           &   ! surface pressure
       geos_heat_flux_sfc,    &   ! geostrophic surface pot. temp. flux  (times f)
       spv_flux_y_avg,        &   ! y component of surface PV flux
       bt_sfctn_avg,          &   ! barotropic streamfunction 
       bt_sfctn_var_avg,      &   ! barotropic streamfunction variance
       v_baro_avg,            &   ! barotropic meridional velocity
       u_baro_avg,            &   ! barotropic zonal velocity
       v_baro_var_avg,        &   ! variance of barotropic meridional velocity
       u_baro_var_avg,        &   ! variance of barotropic zonal velocity
       theta_s_mont_x,        &   !
       precip_cond_avg,       &   ! large scale cond precipiation
       precip_conv_avg,       &   ! convection precipiation
       precip_intensity,      &   ! precipiation intensity
       precip_cond_intensity, &   ! large scale cond precipiation intensity
       precip_conv_intensity, &   ! convection precipiation intensity
       wet_days,              &   ! count of wet days
       wet_days_cond,         &   ! count of wet days from large scale cond
       wet_days_conv              ! count of wet days from convection

! ---------- zonally averaged gridpoint variables on theta surfaces  ----------

  real, allocatable, dimension(:, :) ::                                       &
       sfctn_isavg,        &   ! streamfunction
       ipv_isavg,          &   ! \avg{density * IPV}
       ipv2_isavg,         &   ! isentropic mean of squared potential vorticity
       ipv_flux_isavg,     &   ! zonal mean IPV flux
       p_isavg,            &   ! pressure
       p_var_isavg,        &   ! variance of pressure
       dens_isavg,         &   ! density
       exner_isavg,        &   ! exner function
       pexner_isavg,       &   ! pressure * (exner function)
       sfc_frequency,      &   ! frequency distribution of surface temperatures
       v_isavg,            &   ! meridional wind
       u_isavg,            &   ! zonal wind
       shum_isavg,         &   ! specific humidity
       ozone_isavg,        &   ! ozone
       rhum_isavg,         &   ! relative humidity
       sat_shum_isavg,     &   ! saturation specific humidity
       vshum_isavg,        &   ! specific humidity * meridional velocity
       vozone_isavg,       &   ! ozone * meridional velocity
       vsat_shum_isavg,    &   ! saturation specific humidity * meridional velocity
       v2_isavg,           &   ! (meridional wind)^2
       u2_isavg,           &   ! (zonal wind)^2
       rho_mont_x_isavg,   &   ! density * x deriv. of Montgomery stfcn
       mont_isavg,         &   ! montgomery streamfunction
       mont_var_isavg,     &   ! variance of montgomery streamfunction
       mont_x_isavg            ! x derivative of montgomery streamfunction

  real, allocatable, dimension(:, :) ::                                       &
       v_theta_isavg,      &   ! vertical derivative of pressure on theta surf.
       v_isavg_nw,         &   ! the not-weighted v on theta surfaces
       pv_isavg,           &   ! \overline{v*pressure}
       pv_eddy_isavg,      &   ! pv_isavg-\overline{v}\overline{p}
       pv_theta_isavg,     &   ! \overline{v*pressure_theta}
       pv_theta_eddy_isavg,&   ! pv_theta_isavg-\overline{v*pressure_theta}
       rhov_isavg,         &   ! rho_theta*v (mass flux)
       rhov_eddy_isavg         ! rhov_isavg-\overline{rho_theta*v} 

! ---------------------------- namelist variables -----------------------------

  integer, parameter ::                                                       &
       max_len = 300           ! maximum length of namelist string variables
                    
  integer ::                                                                  &
       DataIn,             &   ! input momentum fields (u_and_v or vor_and_div)
       data_source,        &   ! source of data (FMS=1, NCEP=2, CCM3=3)
       num_fourier,        &   ! number of Fourier waves
       num_segments = 1,   &   ! number of segments (e.g., pentads)        
       num_bin,            &   ! number of bins for relative humidity pdf
       MaxIsentrLev            ! number of isentropic levels

  logical ::                                                                  &
       smooth_surface,     &   ! include topography
       is_gaussian,        &   ! .true. for data on Gaussian grid
       moisture,           &   ! flag to analyze moisture
       moist_tend,         &   ! flag to analyze moist tendencies
       precip,             &   ! flag to analyze precipitation
       ozone,              &   ! flag to analyze ozone
       virtual,            &   ! flag for virtual temp effect for z and svol
       moist_isentropes,   &   ! flag to analyze data on moist isentropes
       use_linearCC            ! flag to use linearized CC  

  real ::                                                                     &
       TimeIn = 0.,        &   ! start time of analysis       
       PotTempMin,         &   ! minimum isentropic level
       PotTempMax,         &   ! maximum isentropic level
       delta_t,            &   ! timestep (seconds) for diabatic diagnostics
       roughness,          &   ! roughness length for momentum
       precip_threshold        ! threshold for wet days

  character (len = max_len) ::                                                &
       UInputFileName,            &   ! zonal wind input file
       VInputFileName,            &   ! meridional wind input file
       VORInputFileName,          &   ! vorticity input file
       DIVInputFileName,          &   ! divergence input file 
       TEMPInputFileName,         &   ! temperature input file 
       PSInputFileName,           &   ! surface pressuse input file
       SGInputFileName,           &   ! surface geopotential input file
       ShumInputFileName,         &   ! specific humidity input file
       CondInputFileName,         &   ! large scale condensation tendency input file 
       ConvInputFileName,         &   ! convection tendency input file 
       PrecipCondInputFileName,   &   ! large scale condensation precipitation input file 
       PrecipConvInputFileName,   &   ! convection precipitation nput file 
       UVarName,                  &   ! zonal wind variable name
       VVarName,                  &   ! meridional wind variable name
       VORVarName,                &   ! vorticity variable name
       DIVVarName,                &   ! divergence variable name
       TEMPVarName,               &   ! temperature variable name
       PSVarName,                 &   ! surface pressure variable name
       SGVarName,                 &   ! surface geopotential variable name
       ShumVarName,               &   ! specific humidity variable name
       CondVarName,               &   ! large scale condensation tendency variable name
       ConvVarName,               &   ! convection tendency variable name
       PrecipCondVarName,         &   ! large scale condensation precipitation variable name
       PrecipConvVarName,         &   ! convection precipitation variable name
       OutputFileName                 ! output file name

  namelist /main_list/                                                        &
                              DataIn,                     data_source,        &
                         num_fourier,                    MaxIsentrLev,        &
                      smooth_surface,                          TimeIn,        &
                          PotTempMin,                      PotTempMax,        &
                            UVarName,                        VVarName,        &
                          VorVarName,                      DivVarName,        &
                         TempVarName,                       PSVarName,        &
                           SGVarName,                                         &
                         ShumVarName,                 SgInputFileName,        &
                         CondVarName,                     ConvVarName,        &
                   PrecipCondVarName,               PrecipConvVarName,        &
                         is_gaussian,                        moisture,        &
                    moist_isentropes,                         num_bin,        &
                               ozone,                      moist_tend,        &
                        use_linearCC,                                         &
                              precip,                precip_threshold,        &
                             virtual,                                         &
                        num_segments


  namelist /filename_list/                                                    &
                      UInputFileName,                  VInputFileName,        &
                    VorInputFileName,                DivInputFileName,        &
                   TempInputFileName,                 PsInputFileName,        &
                   CondInputFileName,               ConvInputFileName,        &
             PrecipCondInputFileName,         PrecipConvInputFileName,        &
                   ShumInputFileName,                  OutputFileName


  contains

!------------------------------------------------------------------------------

    subroutine allocate_and_initialize(num_lat, num_lon, num_lev)

      integer, intent(in) :: num_lat, num_lon, num_lev
      integer             :: i                         ! loop over rhum pdf bins

! ------------------------------ axis variables -------------------------------

      allocate(      deg_lat(num_lat))
      allocate(      sin_lat(num_lat))
      allocate(        sigma(num_lev))
      allocate(     rhum_bin(num_bin))
      allocate(           pk(num_lev+1))
      allocate(           bk(num_lev+1))
      allocate(   cos_lat_xy(num_lon, num_lat))
      allocate(   cos_lat_yz(num_lat, num_lev))

      allocate(  theta_coord(MaxIsentrLev))
      allocate( dtheta_coord(MaxIsentrLev))


! ----------------- gridpoint variables on pressure surfaces ------------------

      allocate(             u_grid(num_lon, num_lat, num_lev))
      allocate(             v_grid(num_lon, num_lat, num_lev))
      allocate(           vor_grid(num_lon, num_lat, num_lev))
      allocate(           div_grid(num_lon, num_lat, num_lev))
      allocate(          temp_grid(num_lon, num_lat, num_lev))
      allocate(  virtual_temp_grid(num_lon, num_lat, num_lev))
      allocate(           pot_temp(num_lon, num_lat, num_lev))
      allocate(            theta_e(num_lon, num_lat, num_lev))
      allocate(        theta_e_sat(num_lon, num_lat, num_lev))
      allocate(          vcos_grid(num_lon, num_lat, num_lev))
      allocate(             w_grid(num_lon, num_lat, num_lev))
      allocate(          w_up_grid(num_lon, num_lat, num_lev))
      allocate(     w_up_freq_grid(num_lon, num_lat, num_lev))
      allocate(         sfctn_grid(num_lon, num_lat, num_lev+1))
      allocate(     hor_sfctn_grid(num_lon, num_lat, num_lev))
      allocate(             p_half(num_lon, num_lat, num_lev+1))
      allocate(             p_full(num_lon, num_lat, num_lev))
      allocate(          shum_grid(num_lon, num_lat, num_lev))
      allocate(    dt_qg_cond_grid(num_lon, num_lat, num_lev))
      allocate(    dt_qg_conv_grid(num_lon, num_lat, num_lev))
      allocate( dt_qg_cond_counter(num_lon, num_lat, num_lev))
      allocate( dt_qg_conv_counter(num_lon, num_lat, num_lev))
      allocate(        sat_mr_grid(num_lon, num_lat, num_lev))
      allocate(      sat_shum_grid(num_lon, num_lat, num_lev))
      allocate(          rhum_grid(num_lon, num_lat, num_lev))


      allocate( mont_x_grid(num_lon, num_lat, num_lev))
      allocate(   mont_grid(num_lon, num_lat, num_lev))

      allocate(  geopot_grid(num_lon, num_lat, num_lev))

      allocate(            ps_grid(num_lon, num_lat))
      allocate(    surf_montgomery(num_lon, num_lat))
      allocate( dx_surf_montgomery(num_lon, num_lat))
      allocate( dy_surf_montgomery(num_lon, num_lat))
      allocate(     surf_bernoulli(num_lon, num_lat))
      allocate(    dx_surf_pottemp(num_lon, num_lat))
      allocate(    dy_surf_pottemp(num_lon, num_lat))
      allocate(  surf_geopotential(num_lon, num_lat))
      allocate(         spv_flux_y(num_lon, num_lat))
      allocate(      bt_sfctn_grid(num_lon, num_lat))
      allocate(   precip_cond_grid(num_lon, num_lat))
      allocate(   precip_conv_grid(num_lon, num_lat))
      allocate(  precip_cond_daily(num_lon, num_lat))
      allocate(  precip_conv_daily(num_lon, num_lat))

      allocate(        precip_mask(num_lon, num_lat))

      allocate(          one_array(num_lon, num_lat))

      allocate(    geos_heat_flux_sfc(num_lat))

      allocate(  ucos_zon_spec_avg(num_lat, 0:num_fourier))
      allocate(  vcos_zon_spec_avg(num_lat, 0:num_fourier))
      allocate(  shum_zon_spec_avg(num_lat, 0:num_fourier))

! ------------------- gridpoint variables on theta surfaces -------------------

      allocate(          p_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(       dens_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(          v_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(          u_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(               ipv(num_lon, num_lat, MaxIsentrLev))
      allocate(      sfctn_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(     mont_x_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate( rho_mont_x_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(       mont_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(    abs_vor_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(       shum_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(       rhum_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(   sat_shum_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(      vshum_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(  vsat_shum_isentr(num_lon, num_lat, MaxIsentrLev))
      allocate(             k_sfc(num_lon, num_lat              ))
      allocate(            k_asfc(num_lon, num_lat              ))

! --------- zonally-averaged gridpoint variables on pressure surfaces ---------

      allocate(             p_half_zon(num_lat, num_lev+1))
      allocate(             p_full_avg(num_lat, num_lev  ))
      allocate(                  z_avg(num_lat, num_lev  ))
      allocate(               temp_avg(num_lat, num_lev  ))
      allocate(       virtual_temp_avg(num_lat, num_lev  ))
      allocate(           temp_var_avg(num_lat, num_lev  ))
      allocate(           pot_temp_avg(num_lat, num_lev  ))
      allocate(            theta_e_avg(num_lat, num_lev  ))
      allocate(        theta_e_sat_avg(num_lat, num_lev  ))
      allocate(       pot_temp_var_avg(num_lat, num_lev  ))
      allocate(        theta_e_var_avg(num_lat, num_lev  ))
      allocate(                  u_avg(num_lat, num_lev  ))
      allocate(              u_var_avg(num_lat, num_lev  ))
      allocate(              v_var_avg(num_lat, num_lev  ))
      allocate(              w_var_avg(num_lat, num_lev  ))
      allocate(             w_cube_avg(num_lat, num_lev  ))
      allocate(                  v_avg(num_lat, num_lev  ))
      allocate(                  vort_avg(num_lat, num_lev  ))
      allocate(               vcos_avg(num_lat, num_lev  ))
      allocate(                  w_avg(num_lat, num_lev  ))
      allocate(               w_up_avg(num_lat, num_lev  ))
      allocate(              w_up_freq(num_lat, num_lev  ))
      allocate(              uvcos_avg(num_lat, num_lev  ))
      allocate(                 wu_avg(num_lat, num_lev  ))
      allocate(             uvcos_eddy(num_lat, num_lev  ))
      allocate(                wu_eddy(num_lat, num_lev  ))
      allocate(               svol_avg(num_lat, num_lev  ))
      allocate(              sfctn_avg(num_lat, num_lev  ))
      allocate(          hor_sfctn_avg(num_lat, num_lev  ))
      allocate(      hor_sfctn_var_avg(num_lat, num_lev  ))
      allocate(    mrdnl_eddy_heatflux(num_lat, num_lev  ))
      allocate(      mrdnl_eddy_z_flux(num_lat, num_lev  ))
      allocate(      vrtcl_eddy_z_flux(num_lat, num_lev  ))
      allocate(mrdnl_eddy_theta_e_flux(num_lat, num_lev  ))
      allocate(    vrtcl_eddy_heatflux(num_lat, num_lev  ))
      allocate(vrtcl_eddy_theta_e_flux(num_lat, num_lev  ))
      allocate(    mrdnl_eddy_tempflux(num_lat, num_lev  ))
      allocate(    vrtcl_eddy_tempflux(num_lat, num_lev  ))
      allocate(          eddy_temp_div(num_lat, num_lev  ))
      allocate(      d_dy_pot_temp_avg(num_lat, num_lev  ))
      allocate(       d_dy_theta_e_avg(num_lat, num_lev  ))
      allocate(    d_dsig_pot_temp_avg(num_lat, num_lev  ))
      allocate(     d_dsig_theta_e_avg(num_lat, num_lev  ))
      allocate(      d_dp_pot_temp_avg(num_lat, num_lev  ))
      allocate(       d_dp_theta_e_avg(num_lat, num_lev  ))
      allocate(     freq_stat_unstable(num_lat, num_lev  ))
      allocate(     mrdnl_heatflux_avg(num_lat, num_lev  ))
      allocate(       mrdnl_z_flux_avg(num_lat, num_lev  ))
      allocate(       vrtcl_z_flux_avg(num_lat, num_lev  ))
      allocate( mrdnl_theta_e_flux_avg(num_lat, num_lev  ))
      allocate(     vrtcl_heatflux_avg(num_lat, num_lev  ))
      allocate( vrtcl_theta_e_flux_avg(num_lat, num_lev  ))
      allocate(     mrdnl_tempflux_avg(num_lat, num_lev  ))
      allocate(     vrtcl_tempflux_avg(num_lat, num_lev  ))
      allocate(           TEM_res_circ(num_lat, num_lev  ))
      allocate(          MTEM_res_circ(num_lat, num_lev  ))
      allocate(           buoyancy_avg(num_lat, num_lev  ))
      allocate(               shum_avg(num_lat, num_lev  ))
      allocate(         dt_qg_cond_avg(num_lat, num_lev  ))
      allocate(         dt_qg_conv_avg(num_lat, num_lev  ))
      allocate(        dt_qg_cond_freq(num_lat, num_lev  ))
      allocate(        dt_qg_conv_freq(num_lat, num_lev  ))
      allocate(           shum_var_avg(num_lat, num_lev  ))
      allocate(       sat_shum_var_avg(num_lat, num_lev  ))
      allocate(           sat_shum_avg(num_lat, num_lev  ))
      allocate(               rhum_avg(num_lat, num_lev  ))
      allocate(              vshum_avg(num_lat, num_lev  ))
      allocate(              wshum_avg(num_lat, num_lev  ))
      allocate(          temp_shum_avg(num_lat, num_lev  ))

      allocate(           ps_grid_avg(num_lat))
      allocate(        spv_flux_y_avg(num_lat))
      allocate(            u_baro_avg(num_lat))
      allocate(            v_baro_avg(num_lat))
      allocate(        u_baro_var_avg(num_lat))
      allocate(        v_baro_var_avg(num_lat))
      allocate(        theta_s_mont_x(num_lat))
      allocate(          bt_sfctn_avg(num_lat))
      allocate(      bt_sfctn_var_avg(num_lat))
      allocate(       precip_cond_avg(num_lat))
      allocate(       precip_conv_avg(num_lat))
      allocate(      precip_intensity(num_lat))
      allocate( precip_cond_intensity(num_lat))
      allocate( precip_conv_intensity(num_lat))
      allocate(              wet_days(num_lat))
      allocate(         wet_days_cond(num_lat))
      allocate(         wet_days_conv(num_lat))

! ---------- zonally averaged gridpoint variables on theta surfaces -----------

      allocate(      sfctn_isavg(num_lat, MaxIsentrLev))
      allocate(        ipv_isavg(num_lat, MaxIsentrLev))
      allocate(       ipv2_isavg(num_lat, MaxIsentrLev))
      allocate(   ipv_flux_isavg(num_lat, MaxIsentrLev))
      allocate(          p_isavg(num_lat, MaxIsentrLev))
      allocate(      p_var_isavg(num_lat, MaxIsentrLev))
      allocate(       dens_isavg(num_lat, MaxIsentrLev))
      allocate(      exner_isavg(num_lat, MaxIsentrLev))
      allocate(     pexner_isavg(num_lat, MaxIsentrLev))
      allocate(          v_isavg(num_lat, MaxIsentrLev))
      allocate(          u_isavg(num_lat, MaxIsentrLev))
      allocate(         v2_isavg(num_lat, MaxIsentrLev))
      allocate(         u2_isavg(num_lat, MaxIsentrLev))
      allocate(    sfc_frequency(num_lat, MaxIsentrLev))
      allocate(     mont_x_isavg(num_lat, MaxIsentrLev))
      allocate(       mont_isavg(num_lat, MaxIsentrLev))
      allocate(   mont_var_isavg(num_lat, MaxIsentrLev))
      allocate( rho_mont_x_isavg(num_lat, MaxIsentrLev))

      allocate(         v_theta_isavg(num_lat, MaxIsentrLev))
      allocate(            v_isavg_nw(num_lat, MaxIsentrLev))
      allocate(              pv_isavg(num_lat, MaxIsentrLev))
      allocate(         pv_eddy_isavg(num_lat, MaxIsentrLev))
      allocate(        pv_theta_isavg(num_lat, MaxIsentrLev))
      allocate(   pv_theta_eddy_isavg(num_lat, MaxIsentrLev))
      allocate(            rhov_isavg(num_lat, MaxIsentrLev))
      allocate(       rhov_eddy_isavg(num_lat, MaxIsentrLev))
      allocate(            shum_isavg(num_lat, MaxIsentrLev))
      allocate(            rhum_isavg(num_lat, MaxIsentrLev))
      allocate(        sat_shum_isavg(num_lat, MaxIsentrLev))
      allocate(           vshum_isavg(num_lat, MaxIsentrLev))
      allocate(       vsat_shum_isavg(num_lat, MaxIsentrLev))

      allocate(  pot_temp_spec( num_fourier+1, num_fourier+1, num_lev))
      allocate(   theta_e_spec( num_fourier+1, num_fourier+1, num_lev))
      allocate(      conv_spec( num_fourier+1, num_fourier+1, num_lev))
      allocate(      shum_spec( num_fourier+1, num_fourier+1, num_lev))
      allocate(  sat_shum_spec( num_fourier+1, num_fourier+1, num_lev))
      allocate(      rhum_spec( num_fourier+1, num_fourier+1, num_lev))
      allocate(  bc_sfctn_spec( num_fourier+1, num_fourier+1))
      allocate(  bt_sfctn_spec( num_fourier+1, num_fourier+1))

      allocate(       rhum_pdf(num_lat, num_lev, num_bin))

      if(ozone) then
         allocate(         ozone_grid( num_lon, num_lat, num_lev ))
         allocate(       ozone_isentr( num_lon, num_lat, MaxIsentrLev ))
         allocate(      vozone_isentr( num_lon, num_lat, MaxIsentrLev ))
         allocate(          ozone_avg( num_lat, num_lev ))
         allocate(         vozone_avg( num_lat, num_lev ))
         allocate(        ozone_isavg( num_lat, MaxIsentrLev ))
         allocate(       vozone_isavg( num_lat, MaxIsentrLev ))
         allocate(         ozone_spec( num_fourier+1, num_fourier+1, num_lev ))
         allocate( ozone_zon_spec_avg( num_lat, 0:num_fourier ))
         ozone_spec          =  0.0
         ozone_zon_spec_avg  =  0.0
         ozone_avg           =  0.0
         ozone_isavg         =  0.0
         vozone_avg          =  0.0
         vozone_isavg          =  0.0
      endif


      v_theta_isavg           =  0.0
      v_isavg_nw              =  0.0
      pv_isavg                =  0.0
      pv_eddy_isavg           =  0.0
      pv_theta_eddy_isavg     =  0.0
      pv_theta_isavg          =  0.0
      rhov_isavg              =  0.0
      rhov_eddy_isavg         =  0.0

      p_full_avg              =  0.0
      bt_sfctn_avg            =  0.0
      bt_sfctn_var_avg        =  0.0
      z_avg                   =  0.0
      temp_avg                =  0.0
      virtual_temp_avg        =  0.0
      temp_var_avg            =  0.0
      pot_temp_avg            =  0.0
      theta_e_avg             =  0.0
      theta_e_sat_avg         =  0.0
      pot_temp_var_avg        =  0.0
      theta_e_var_avg         =  0.0
      u_avg                   =  0.0
      spv_flux_y_avg          =  0.0
      v_avg                   =  0.0
      vort_avg                =  0.0
      u_var_avg               =  0.0
      v_var_avg               =  0.0
      w_var_avg               =  0.0
      w_cube_avg              =  0.0
      vcos_avg                =  0.0
      w_avg                   =  0.0
      w_up_avg                =  0.0
      w_up_freq               =  0.0
      uvcos_avg               =  0.0
      wu_avg                  =  0.0
      svol_avg                =  0.0
      sfctn_avg               =  0.0
      hor_sfctn_avg           =  0.0
      hor_sfctn_var_avg       =  0.0
      ps_grid_avg             =  0.0
      sfctn_isavg             =  0.0
      ipv_isavg               =  0.0
      ipv2_isavg              =  0.0
      ipv_flux_isavg          =  0.0
      p_isavg                 =  0.0
      p_var_isavg             =  0.0
      exner_isavg             =  0.0
      pexner_isavg            =  0.0
      dens_isavg              =  0.0
      v_isavg                 =  0.0
      u_isavg                 =  0.0
      v2_isavg                =  0.0
      u2_isavg                =  0.0
      sfc_frequency           =  0.0
      d_dy_pot_temp_avg       =  0.0
      d_dy_theta_e_avg        =  0.0
      d_dsig_pot_temp_avg     =  0.0
      d_dsig_theta_e_avg      =  0.0
      d_dp_pot_temp_avg       =  0.0
      d_dp_theta_e_avg        =  0.0
      freq_stat_unstable      =  0.0
      mrdnl_heatflux_avg      =  0.0
      mrdnl_z_flux_avg        =  0.0
      vrtcl_z_flux_avg        =  0.0
      mrdnl_theta_e_flux_avg  =  0.0
      vrtcl_heatflux_avg      =  0.0
      vrtcl_theta_e_flux_avg  =  0.0
      mrdnl_tempflux_avg      =  0.0
      vrtcl_tempflux_avg      =  0.0
      vrtcl_eddy_heatflux     =  0.0
      vrtcl_eddy_theta_e_flux =  0.0
      mrdnl_eddy_heatflux     =  0.0
      mrdnl_eddy_z_flux       =  0.0
      vrtcl_eddy_z_flux       =  0.0
      mrdnl_eddy_theta_e_flux =  0.0
      vrtcl_eddy_tempflux     =  0.0
      mrdnl_eddy_tempflux     =  0.0
      eddy_temp_div           =  0.0
      ucos_zon_spec_avg       =  0.0
      vcos_zon_spec_avg       =  0.0
      shum_zon_spec_avg       =  0.0
      buoyancy_avg            =  0.0
      theta_s_mont_x          =  0.0
      TEM_res_circ            =  0.0
      MTEM_res_circ           =  0.0

      shum_avg                =  0.0
      shum_var_avg            =  0.0
      dt_qg_cond_avg          =  0.0
      dt_qg_conv_avg          =  0.0
      dt_qg_cond_freq         =  0.0
      dt_qg_conv_freq         =  0.0
      sat_shum_var_avg        =  0.0
      sat_shum_avg            =  0.0
      rhum_avg                =  0.0
      shum_isavg              =  0.0
      rhum_isavg              =  0.0
      sat_shum_isavg          =  0.0
      vshum_avg               =  0.0
      wshum_avg               =  0.0
      temp_shum_avg           =  0.0
      vshum_isavg             =  0.0
      vsat_shum_isavg         =  0.0
      precip_cond_avg         =  0.0
      precip_conv_avg         =  0.0
      precip_intensity        =  0.0
      precip_cond_intensity   =  0.0
      precip_conv_intensity   =  0.0
      wet_days                =  0.0
      wet_days_cond           =  0.0
      wet_days_conv           =  0.0
      
      p_isentr                =  0.0
      dens_isentr             =  0.0
      v_isentr                =  0.0
      u_isentr                =  0.0
      ipv                     =  0.0
      sfctn_isentr            =  0.0
      abs_vor_isentr          =  0.0
      k_sfc                   =  0
      k_asfc                  =  0
      geos_heat_flux_sfc      =  0.0       
      
      surf_geopotential       =  0.0
      geopot_grid             =  0.0

      u_baro_avg              =  0.0
      v_baro_avg              =  0.0
      u_baro_var_avg          =  0.0
      v_baro_var_avg          =  0.0

      mont_x_isavg            =  0.0
      mont_isavg              =  0.0
      mont_var_isavg          =  0.0
      rho_mont_x_isavg        =  0.0

      bc_sfctn_spec           =  0.0
      bt_sfctn_spec           =  0.0
      pot_temp_spec           =  0.0
      theta_e_spec            =  0.0
      conv_spec               =  0.0
      shum_spec               =  0.0
      sat_shum_spec           =  0.0
      rhum_spec               =  0.0

      precip_cond_daily       =  0.0
      precip_conv_daily       =  0.0

      one_array               =  1.0
   
      rhum_pdf                =  0.0 

      ! pdf bins for relative humidity
      ! include one negative bin
      do i=1, num_bin
         rhum_bin(i) =  (float(i) - 1.5) / (float(num_bin)-1)
      enddo


     end subroutine allocate_and_initialize

   end module vars
